local Block = require("blocks.Block")
local EventDispatcher = require("events.EventDispatcher")
local Noise = require("generation.noise.Noise")
local materials = require("material")
local blocks = require("blocks.blocks")
local points = require("point")
local Point = points.Point
local deltas = points.deltas
local debugUtil = require("debug.debugUtil")
local Generator = require("generation.generators.Generator")
local Chunk = require("game.Chunk")
local Region = require("game.Region")
local LiquidHelper = require("physics.LiquidHelper")
local idTable = require("game.IdTable"):getInstance()

local chunkPool = {}
local regionPool = {}

local function make2dArray(width,height, constructor)
  local result = {}
  for y=1,height do
    local row = {}
    result[y] = row
    for x=1,width do
      row[x] = constructor(x,y)
      if row[x]==nil then
        error("constructor was faulty!")
      end
    end
  end
  return result
end








local BlockLayer = {}
function BlockLayer:new(width, height, seed, index)
  self.__index = self
  if not index then
    index = 0
  end
  local instance = {
    width=width,
    height=height,
    groundBlockLayer=math.ceil(height/2),
    toUpdate={},
    manyUpdatesCheck = 0,
    generatorSeed = seed,
    chunkSize = 16,
    chunkLoadRadius = 4,
    regionLoadRadius = 1,
    chunkLoaders = {},
    chunks = {},
    regions = {},
    index = index
  }

  setmetatable(instance, self)

  instance.generatorThread = love.thread.newThread("threads/generatorThread.lua")

  instance.generatorParamsChannel = love.thread.getChannel("generatorParams"..instance.generatorSeed)
  instance.generatorThread:start(idTable:toThreadData(), instance.generatorSeed, instance.groundBlockLayer, instance.height)
  local threadError = instance.generatorThread:getError()
  if threadError then
    error("generator thread: "..threadError)
  end


  instance.liquidHelper = LiquidHelper:new(instance)

  instance.defaultBlock = blocks.rawnew("core:grass")
  instance.defaultBlock.destructible = false
  --instance:generateBlocks()
  EventDispatcher:new(instance)

  return instance
end

function BlockLayer:handleChunkLoaders()

  local start = love.timer.getTime()
  local toLoad = {}
  local regionIdsToLoad = {}

  local count = 0

  for _, chunkLoader in ipairs(self.chunkLoaders) do
    local pos = chunkLoader.position:clone()
    if not chunkLoader.isBlock then
      --print("pos "..tostring(pos).. "becomes... ")
      pos = (pos * (1/BLOCK_SIZE)):round()
      --print("    ..."..tostring(pos))
    end
    local r = self.chunkLoadRadius
    local chunkLeft = -r
    local chunkRight = r
    local chunkTop = -r
    local chunkBottom = r


    if chunkLoader.speed then
      local chunkSpeed = (chunkLoader.speed:clone() * (1/(BLOCK_SIZE*self.chunkSize*10))):toLongerInt()
      if chunkSpeed.x > chunkRight then chunkRight = chunkSpeed.x end
      if chunkSpeed.x < chunkLeft  then  chunkLeft = chunkSpeed.x end
      if chunkSpeed.y > chunkBottom then chunkBottom = chunkSpeed.y end
      if chunkSpeed.y <  chunkTop then chunkTop = chunkSpeed.y end
    end


    for dx=chunkLeft,chunkRight do
      for dy=chunkTop,chunkBottom do
        local chunk =  self:chunkAtBlockXY(pos.x+dx*self.chunkSize,pos.y+dy*self.chunkSize)
        toLoad[chunk] = true
      end
    end

    local centralChunk = self:chunkAtBlockXY(pos.x, pos.y)
    local centralRegionX = Region.getXForChunk(centralChunk)
    local centralRegionY = Region.getYForChunk(centralChunk)
    r = self.regionLoadRadius
    for dx=-r,r do
      for dy=-r,r do
        regionIdsToLoad[ Region.getIdXYI(centralRegionX+dx, centralRegionY+dy, self.index)] = true
      end
    end

  end


  for key, region in pairs(self.regions) do
    if not regionIdsToLoad[key] then
      region:release()
      if not region.isReleasing then
        regionPool[#regionPool+1] = region
        self.regions[key] = nil

      end
    end
  end

  for regionKey, _ in pairs(regionIdsToLoad) do
    if not self.regions[regionKey] then
      local region
      if #regionPool>0 then
        region = regionPool[#regionPool]
        region:initFromKey(regionKey)
        regionPool[#regionPool]=nil

      else
        region = Region.createByKey(regionKey)
      end
      self.regions[regionKey] = region
      region:init()
    end
  end


  local toUnload = {}
  for key, chunk in pairs(self.chunks) do
    if not toLoad[chunk] then
      self:clearChunk(chunk)
      count = count + 1
    end
  end

  for chunk, _ in pairs(toLoad) do
    if not chunk.isLoaded then
      local region = self.regions[Region.getIdForChunk(chunk)]
      if region then
        region:loadChunk(chunk)
      end
      count = count + 1
    end
  end

  if count > 0 then
    local delta = (love.timer.getTime() - start)
    if delta > 0.1 then
      print("chunk operations = "..tostring(count).." times "..delta)
    end
  end

end

local function keyf(x,y)
  return x*100000+ y
end

function BlockLayer:chunkAtBlockXY(x,y)
  x = x - x % self.chunkSize
  assert(y % self.chunkSize >= 0)
  y = y - y % self.chunkSize

  local key = keyf(x,y)

  if not self.chunks[key] then
    local chunk
    if #chunkPool > 0 then
      chunk = chunkPool[#chunkPool]
      chunk.fromPoint.x = x
      chunk.fromPoint.y = y
      chunk.toPoint.x = x+self.chunkSize-1
      chunk.toPoint.y = y+self.chunkSize-1
      chunk.layer = self
      chunkPool[#chunkPool]=nil
    else
      chunk =  Chunk:new(self, points.new(x,y), points.new(x+self.chunkSize-1, y+self.chunkSize-1))
    end


    self.chunks[key]= chunk
  end
  return self.chunks[key]
end

function BlockLayer:chunkAtEntityPos(pt)
  local pos = pt:clone()
  pos = (pos * (1/BLOCK_SIZE)):round()
  return self:chunkAtBlockXY(pos.x, pos.y)
end

function BlockLayer:clearChunk(chunk)
  local key = keyf(chunk.fromPoint.x,chunk.fromPoint.y)
  if chunk.isLoaded then
    local region = self.regions[Region.getIdForChunk(chunk)]
    if region then
      region:saveChunk(chunk)
    end
  end
  if not chunk.isSaving then
    self.chunks[key] = nil
    chunkPool[#chunkPool+1] = chunk
  end
end


function BlockLayer:getBlock(x,y)
  local chunk = self:chunkAtBlockXY(x,y)
  return chunk:getBlock(x,y)
end

function BlockLayer:getBlockPt(pt)
  return self:getBlock(pt.x, pt.y)
end



function BlockLayer:spawnBlock(x,y, block)
  if not block then
    return
  end
  self:chunkAtBlockXY(x,y):setBlock(x,y,block)

  block:setLevel(self)
  block.position.x = x
  block.position.y = y

end

function BlockLayer:setBlock(x,y, block)
  local chunk = self:chunkAtBlockXY(x,y)
  chunk.changed = true

  local oldLevel = block:getLevel()
  if oldLevel then
    oldLevel:setBlock(block.position.x, block.position.y, blocks.new("core:sky"))
  end

  local oldBlock =  chunk:getBlock(x,y)
  if oldBlock then oldBlock:setLevel(nil) end
  chunk:setBlock(x,y,block)
  block.position.x = x
  block.position.y = y
  block:setLevel(self)

  self:scheduleNeighboursForUpdate(block)


  return block
end

function BlockLayer:setBlockPt(pos, block)
  self:setBlock(pos.x, pos.y, block)
end

function BlockLayer:markChunkAtBlockChanged(block)
  self:chunkAtBlockXY(block.position.x,block.position.y).changed = true
end

function BlockLayer:scheduleNeighboursForUpdate(block, radius)
  if not radius then
    radius = 1
  end
  local centerX = block.position.x
  local centerY = block.position.y
  for dx=-radius, radius do
    for dy=-radius, radius do
      self.toUpdate[self:getBlock(centerX+dx, centerY+dy)] = true
    end
  end
end




function BlockLayer:updateBlocks()
  local toUpdate = self.toUpdate
  self.toUpdate = {}
  local priorityKeys = {}
  local priorities = {}
  for block, _ in pairs(toUpdate) do
    local priority = block.updatePriority and block.updatePriority or 0
    if not priorities[priority] then
      priorities[priority] = {}
      table.insert(priorityKeys, priority)
    end
    table.insert(priorities[priority], block)
  end

  table.sort(priorityKeys, function(a,b) return a > b end)
  for _, priority in ipairs(priorityKeys) do
    for __, block in ipairs(priorities[priority]) do
      block:update()
    end
  end

  --[[self.manyUpdatesCheck = self.manyUpdatesCheck + 1
  for block, _ in pairs(toUpdate) do
    block:update()
  end]]--

end

function BlockLayer:update()
  self:handleChunkLoaders()
  self:updateBlocks()
end


function BlockLayer:dropBlock(block)
  self:setBlock(block.position.x, block.position.y, blocks.new("core:sky"))
  self:dispatchEvent("dropBlock", block)
end



function BlockLayer:getBlockPx(pixelPos)
  local blockX = math.floor(pixelPos.x / BLOCK_SIZE)
  local blockY = math.floor(pixelPos.y / BLOCK_SIZE)
  --print("xy "..blockX.." "..blockY)
  return self:getBlock(blockX, blockY)
end

return BlockLayer
