local ffi = require("ffi")
ffi.cdef[[
  void* malloc(size_t memblock);
  void free(void* memblock);
]]


local _blocks = require("blocks.blocks")
local threadCallbacks = require("threads.ThreadCallBacks").getInstance()
local stringutils = require("utils.stringutils")

local fileutils = require("utils.fileutils")
local points = require("point")
local idTable = require("game.IdTable").getInstance()

local blocksTablesCache = {}


local Chunk = {}
Chunk.totalSaving = 0
Chunk.totalLoading = 0
Chunk.saveCount = 0
Chunk.loadCount = 0

function Chunk.getStats()
  local result = ""

  if Chunk.saveCount > 0 then
    result = result .. "saves=".. tostring(Chunk.totalSaving / Chunk.saveCount)
  else
    result = result .. "no saves"
  end
  result = result.."\t\t\t"
  if Chunk.loadCount > 0 then
    result = result .. "loads="..tostring(Chunk.totalLoading / Chunk.loadCount)
  else
    result = result .. "no loads"
  end
  return result
end
function Chunk:new(layer, fromPoint, toPoint)
  self.__index = self
  local instance = {layer=layer, fromPoint=fromPoint,toPoint=toPoint, isLoaded=false, isLoading = false, isSaving = false}
  setmetatable(instance,self)
  return instance
end



function Chunk:unload()
  local start = love.timer.getTime()
  if not self.isLoaded then
    return
  end

  local blocks = self.blocks
  for i=1,#blocks do
    blocks[i]:setLevel(nil)
    _blocks.release(blocks[i])
    blocks[i] = nil
  end
  self.blocks = nil
  blocksTablesCache[#blocksTablesCache+1] = blocks

  self.isLoaded = false
  self.isLoading = false
  self.changed = true

  Chunk.saveCount = Chunk.saveCount + 1
  Chunk.totalSaving = Chunk.totalSaving + love.timer.getTime() - start
end

function Chunk:getId()
  return "chunk"..tostring(self.fromPoint.x).."_"..tostring(self.fromPoint.y).."_"..tostring(self.toPoint.x).."_"..tostring(self.toPoint.y).."("..tostring(self.layer.index)..")"
end

function Chunk:getFileName()
  return fileutils.getChunkFileNameById(self:getId())
end

function Chunk:getDataSize()
  local saveSize = 2 * self:getBlocksCount()
  local memSize = ffi.sizeof("double")*saveSize
  return memSize
end



function Chunk:getBlocksCount()
  local fromPoint = self.fromPoint
  local toPoint = self.toPoint
  return(toPoint.x- fromPoint.x +1 ) * (toPoint.y - fromPoint.y + 1)
end

function Chunk:getData()
  local fromPoint = self.fromPoint
  local toPoint = self.toPoint


  local saveSize = 2 * (toPoint.x- fromPoint.x +1 ) * (toPoint.y - fromPoint.y + 1)
  local memSize = ffi.sizeof("double")*saveSize
  local result =ffi.cast("double *",  ffi.C.malloc(memSize))

  local nextIndex = 0

  for y=fromPoint.y, toPoint.y do
    for x=fromPoint.x, toPoint.x do
      local block = self.layer:getBlock(x,y)
      assert(idTable:getIndex(block.blockId), " block id not registered! "..block.blockId)
      result[nextIndex] = idTable:getIndex(block.blockId)
      nextIndex = nextIndex + 1
      result[nextIndex] = block:getSaveData()
      nextIndex = nextIndex + 1
    end
  end

  local resultPtr = tonumber(ffi.cast("intptr_t", result))
  local temp = ffi.cast("double*", ffi.cast("intptr_t", resultPtr))
  for i=0, saveSize-1 do
    if i%2==0 and temp[i] ~= math.floor(temp[i]) then
      error("detetcted invalid save!")
    end
  end
  return resultPtr
end

function Chunk:onSaved()
  self:unload()
  self.isSaving = false
end


function Chunk:generate()
  if not self.layer.generatorThread:isRunning() then
    local lastError = self.layer.generatorThread:getError()
    error("generator thread not running: "..lastError)
  end
  self.layer.generatorParamsChannel:push({self.layer.index, self.fromPoint.x, self.fromPoint.y, self.toPoint.x, self.toPoint.y, threadCallbacks:registerCallBack(function(data)
    self:fromBlockIds(data)
  end)})
end

function Chunk:fromBlockIds(dataPointer)
  local data = ffi.cast("double *", dataPointer)
  local nextIndex = 0
  local blocks = self.blocks
  local block
  for y = self.fromPoint.y, self.toPoint.y do
    for x=self.fromPoint.x, self.toPoint.x do
      block = _blocks.new(idTable:getId(data[nextIndex]))
      blocks[nextIndex+1] = block
      block.position.x = x
      block.position.y = y
      block:setLevel(self.layer)
      nextIndex = nextIndex + 1
    end
  end


  ffi.C.free(data)

  self.isLoaded = true
  self.isLoading = false
  self.changed = true
end



function Chunk:onLoaded(data)
  if #blocksTablesCache > 0 then
    self.blocks = blocksTablesCache[#blocksTablesCache]
    blocksTablesCache[#blocksTablesCache] = nil
  else
    self.blocks =
    {
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,
      nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil,nil
    }
  end
  if data then
    self:loadFromString(data)
    self.isLoaded = true
    self.isLoading = false
    self.changed = true
  else
    self:generate()
  end
end

function Chunk:loadFromString(dataPtr)


  local data = ffi.cast("double *", ffi.cast("intptr_t",dataPtr))

  local start = love.timer.getTime()

  local fromPoint = self.fromPoint
  local toPoint = self.toPoint
  local nextIndex = 0

  local blockId
  local block
  local blockIndex
  local posIndex = 1
  local blocks = self.blocks
  for y=fromPoint.y, toPoint.y do
    for x=fromPoint.x, toPoint.x do
      blockIndex = math.floor(data[nextIndex])
      nextIndex = nextIndex + 1
      blockId = idTable:getId(blockIndex)

      if not blockId then
        blockId = idTable:getId(1)
      end

      block = _blocks.new(blockId)
      blocks[posIndex] = block
      posIndex = posIndex+1
      block.position.x = x
      block.position.y = y
      block:setLevel(self.layer)
      block:fromSaveData(data[nextIndex])
      nextIndex = nextIndex + 1
    end
  end

  ffi.C.free(data)

  Chunk.loadCount = Chunk.loadCount + 1
  Chunk.totalLoading = Chunk.totalLoading + love.timer.getTime() - start
end

function Chunk:getChunkX()
  return self.fromPoint.x / self.layer.chunkSize
end

function Chunk:getChunkY()
  return self.fromPoint.y / self.layer.chunkSize
end

function Chunk:getBlock(x,y)
  local dx = x-  self.fromPoint.x
  local dy = y - self.fromPoint.y
  if dx < 0 or dx > 15 or dy < 0 or dy > 15 then
    error("requesting from wrong chunk!")
  end
  if self.isLoaded then
    local block =  self.blocks[1 + 16*dy + dx]
    if not block then error("self blocks not filled!") end
    return block
  else
    return self.layer.defaultBlock
  end
end

function Chunk:setBlock(x,y, block)
  if self.isLoaded then
    self.blocks[1+ (x-self.fromPoint.x)+16*(y-self.fromPoint.y)] = block
  end
end

return Chunk
