local EntityMovement = require("physics.EntityMovement")
local points = require("point")
local Point = points.Point
local SurfaceGravity = require("physics.SurfaceGravity")
local SurfaceResistance = require("physics.SurfaceResistance")
local BlockCollider = require("physics.BlockCollider")
local BlockEntity = require("entities.BlockEntity")

local GameArea = {}
GameArea.LEVEL_WIDTH = 1000
GameArea.LEVEL_HEIGHT = 500

function GameArea:new()
  self.__index = self
  local topLeft = points.new(0,0)
  local bottomRight = points.new(self.LEVEL_WIDTH, self.LEVEL_HEIGHT) * BLOCK_SIZE
  assert(SurfaceGravity)
  local instance = {
    entities = {},
    layers = {},
    limits = {
      topLeft = topLeft,
      bottomRight = bottomRight
    },
    physics = {
        movement = EntityMovement:new(topLeft,bottomRight),
        gravity = SurfaceGravity:new(),
        blockCollider = BlockCollider:new(),
    },
    mainLayer= nil
  }
  instance.dropListener = function(event, data)
    instance:dropBlock(data)
  end

  return setmetatable(instance,self)
end

function GameArea:addLayer(layer, main)
  table.insert(self.layers, layer)
  if main then
    self.mainLayer = layer
    self.physics.blockCollider:setLayer(layer)
    self.physics.surfaceResistance = SurfaceResistance:new(layer)
    layer:addEventListener("dropBlock", self.dropListener)
  end
end

function GameArea:dropBlock(block)
  if block.getDrops then
    for _, entity in ipairs(block:getDrops()) do
      self:addEntity(entity)
    end
  else
    local blockEntity = BlockEntity:new(block)
    self:addEntity(blockEntity)
  end
end


function GameArea:addEntity(entity)
  table.insert(self.entities, entity)
end

function GameArea:removeEntity(entity)
  local toRemove = {}
  for i, v in ipairs(self.entitites) do
    if entity == v then
      table.insert(toRemove, i)
    end
  end

  for i = #toRemove, 1, -1 do
    local indexToRemove = toRemove[i]
    table.remove(self.entities, indexToRemove)
  end
end

function GameArea:handleDt(dt)
  self.dt = dt
  self:handleBlocks()

  self:handleEntities()

end

function GameArea:handleEntities()
  local entities = self.entities
  if not self.times then
    self.times = {0,0,0,0}
  end
  local times = self.times
  local previousTime = love.timer.getTime()
  local function addTime(i)
    times[i] = times[i] + love.timer.getTime() - previousTime
    previousTime = love.timer.getTime()
  end
  for i, entity in ipairs(entities) do
    local chunk = self.mainLayer:chunkAtEntityPos(entity.position)
    local moving
    if chunk.isLoaded then
      self.physics.gravity:addForce(entity)
      addTime(1)
      entity:applyForces()
      moving = not self.physics.surfaceResistance:handle(entity)
      addTime(2)
      if moving then self.physics.movement:handle(entity, self.dt) end
      addTime(3)
      if moving then self.physics.blockCollider:handle(entity) end
      addTime(4)
    end

  end
  self.timesFormatted = table.concat(self.times, "\t")

end

function GameArea:getThingsInArea(topLeft, size, layer)
  local usedBlocks= {}
  local count = 0
  local layersToCheck = layer and {layer} or self.layers
  for _, layer in ipairs(layersToCheck) do
    for x = topLeft.x, topLeft.x+size.x, BLOCK_SIZE do
      for y = topLeft.y, topLeft.y+size.y, BLOCK_SIZE do
        usedBlocks[layer:getBlockPx(points.new(x,y))] = true
        count = count + 1
      end
    end
    if count > 0 then
      break
    end
  end
  local blocksArray = {}
  for k,v in pairs(usedBlocks) do
    table.insert(blocksArray, k)
  end
  return blocksArray
end

function GameArea:applyTool(tool)
  local topLeft = tool.position + tool.shift
  local things = self:getThingsInArea(topLeft, tool.targetSize, self.mainLayer)
  tool:apply(things)
end

function GameArea:handleBlocks()
  for _, layer in ipairs(self.layers) do
    layer:update()
  end
end

return GameArea
