local ffi = require("ffi")
ffi.cdef[[
  void* malloc(size_t memsize);
  void free(void* memblock);
]]
local fileutils = require("utils.fileutils")
local SAVE_DIRECTORY = love.filesystem.getSaveDirectory()
-- will handle large files with several chunks
local Region = {}
Region.size = 32
Region.maxChunks = Region.size * Region.size


function Region.getIdXYI(x,y, index)
  return x .. "_"..y.."_"..index
end

function Region.getXForChunk(chunk)
  return  math.floor(chunk:getChunkX() / Region.size)
end

function Region.getYForChunk(chunk)
  return math.floor(chunk:getChunkY()/ Region.size)
end

function Region.getIdForChunk(chunk)
  return Region.getIdXYI(Region.getXForChunk(chunk),Region.getYForChunk(chunk),chunk.layer.index)
end

function Region.createByKey(key)
  local nextPart = key:gmatch("[^_]+")
  local nextNum = function() return tonumber(nextPart()) end
  return Region:new(nextNum(), nextNum(), nextNum())
end

function Region:initFromKey(key)
  local nextPart = key:gmatch("[^_]+")
  local nextNum = function() return tonumber(nextPart()) end
  self.x = nextNum()
  self.y = nextNum()
  self.index = nextNum()
end

function Region:new(x, y, index)
  self.__index = self
  local instance = {isLoading=false, isLoaded=false, isSaving=false, offsets={}, x=x, y=y, index=index, fileHandle=nil}

  setmetatable(instance, self)
  return instance
end

function Region:getId()
  return Region.getIdXYI(self.x,self.y,self.index)
end


function Region:getFileName()
  return self:getId()..".rgn"
end

function Region:getFullFileName()
  return SAVE_DIRECTORY.."/"..fileutils.getRegionsRoot().."/"..self:getFileName()
end

function Region:init()
  if self.isLoading or self.isLoaded then
    return
  end
  self.isLoading = true
  fileutils.openFileAsync(self:getFullFileName(), function(result)
    self.fileHandle = result
    self:readHeader()
  end)
end

function Region:readHeader()
  for i=1, self.maxChunks do
    self.offsets[i] = -1
  end

  local dataSize = ffi.sizeof("double")*self.maxChunks
  assert(dataSize)
  --err("requesting header")
  fileutils.readFileAsync(self.fileHandle, 0, dataSize, function(data,data2)
    --err("handling!!!")
    if data2 then
      local ptr = data
      ffi.C.free(ffi.cast("char*", ptr))
      --err("nothing in file")
      --read not successful
      self:writeHeader()
    else
      --err("something in file")
      local dataPtr = ffi.cast("double*", data)

      local maxOffset = dataSize
      for i=1,self.maxChunks do
        self.offsets[i] =  dataPtr[i-1]
      end
      self.isLoading = false
      self.isLoaded = true
    end
  end)
end

function Region:writeHeader()
  local dataSize = ffi.sizeof("double")*self.maxChunks
  local dataPtr = ffi.cast("double *", ffi.C.malloc(dataSize))
  for i=1, self.maxChunks do
    if self.offsets[i] then
      dataPtr[i-1] = self.offsets[i]
    end
  end
  fileutils.writeFileAsync(self.fileHandle, 0, tonumber(ffi.cast("intptr_t", dataPtr)), dataSize, function(result)
    self.headerReady = true
    self.isLoading = false
    self.isLoaded = true
  end)
end

function Region:getChunkIndex(chunk)
  return 1+ (chunk:getChunkY() % self.size)*self.size + chunk:getChunkX() % self.size
end

function Region:saveChunk(chunk)
  if not self.isLoaded then return end
  if chunk.isSaving then return end
  chunk.isSaving = true

  local index = self:getChunkIndex(chunk)
  local offset = self.offsets[index]

  fileutils.writeFileAsync(self.fileHandle, offset, chunk:getData(), chunk:getDataSize(), function(result)
    self.offsets[index] = result
    chunk:onSaved()
  end)


end

function Region:loadChunk(chunk)
  if not self.isLoaded then
    return
  end
  if chunk.isLoading then
    return
  end
  chunk.isLoading = true
  local index = self:getChunkIndex(chunk)
  local offset = self.offsets[index]

  if offset == -1 then
    chunk:onLoaded(nil)
  else
    fileutils.readFileAsync(self.fileHandle, offset, chunk:getDataSize(), function(result)
      chunk:onLoaded(result)
    end)
  end
end

function Region:release()
  if self.isReleasing or self.isLoading or not self.isLoaded then return end
  self.isReleasing = true
  self:writeHeader()
  fileutils.closeFileAsync(self.fileHandle, function(result)
    self.fileHandle=nil
    self.isReleasing = false
    self.isLoaded = false
    self.isLoading = false
  end)
end

return Region
