local Rng = require("generation.rng.Rng")
local Noise = {}
Noise.seed = 925960
Noise.randomSize = 256
local debugUtil = require("debug.debugUtil")

function Noise:new(seed, randomSize)
  self.__index = self
  if seed then
    print("INIT WITH SEED")
  end
  local instance = {seed=seed, randomSize=randomSize, gradients={}}
  setmetatable(instance, self)
  instance:init()
  return instance
end


function Noise:init()
  self.rng = Rng:new(self.seed)
  print("RNG SEED IS "..tostring(self.seed))
  self.randoms = {}
  for i=1,self.randomSize do
    table.insert(self.randoms, self.rng:next())
  end
end

function Noise:fade(t)
  return t * t * t * ( 10 + t * (-15 + t * 6) )
end

function Noise:rand(index)
  local _i = 1 + math.floor(index % self.randomSize)
  if not _i or not self.randoms[_i] then
    print(_i)
    error("not valid _i "..tostring(_i).. " from index "..tostring(index))
  end
  return self.randoms[_i]
end

function Noise:randomAt(...)
  local arguments = {...}
  if #arguments == 0  then error("at least one coordinate is required!") end
  local result = 0
  for i,v in ipairs(arguments) do

     result = self:rand(v + result)
  end

  return result
end

function Noise:gradientAt1d(x)
  local rand = (-6 + self:randomAt(x)%13 )/6
  if math.abs(rand) > 1 then
    error("> 1 not expected"..tostring(rand))
  end
  return rand
end

function Noise:gradientAtNd(...)
  local args = {...}
  local n = #args
  assert(n>0, "not enough noice coords")
  if n == 1 then
     return self:gradientAt1d(unpack(args))
  end

  local gradients = self:getGradients(n)
  local rand = 1 + self:randomAt(unpack(args)) % #gradients
  return gradients[rand]

end

local function cloneTable(t)
  local res = {}
  for k,v in pairs(t) do
    res[k] = v
  end
  return res
end

function Noise:getGradients(n)

  if self.gradients[n] then
    return self.gradients[n]
  end
  assert(n>0)
  local result

  if n == 1 then
    result = {{1},{0},{-1}}
  else
    local previousLevel = self:getGradients(n-1)
    result = {}
    for newDim=-1,1 do
      for _, entry in ipairs(previousLevel) do
        local entryCopy = cloneTable(entry)
        table.insert(entryCopy, newDim)
        table.insert(result, entryCopy)
      end
    end

  end
  self.gradients[n] = result
  return result


end

function Noise:get1d(x)
  local leftGradient = self:gradientAt1d(math.floor(x))
  local rightGradient = self:gradientAt1d(math.ceil(x))
  local dx = x - math.floor(x)
  local ratio = self:fade(dx)
  local result = leftGradient * dx * (1-ratio)  + rightGradient * (dx - 1) * (ratio)
  return result
end


function Noise:getNd(...)
  debugUtil.resetDelta()
  local args={...}
  if #args == 1 then
    return self:get1d(unpack(args))
  end
  if #args == 2 then
    local x = args[1]
    local y = args[2]
    local dx = x %1
    local dy = y %1
    assert(dx>=0)
    assert(dy>=0)
    debugUtil.addDelta("init")

    local topRight = dx > dy
    local lx = math.floor(x)
    local ly = math.floor(y)
    local rx = math.ceil(x)
    local ry = math.ceil(y)
    local g1 = self:gradientAtNd(lx, ly)
    local g2 = self:gradientAtNd(rx, ry)
    local g3 = self:gradientAtNd(  topRight and rx or lx,  not topRight and ry or ly)
    debugUtil.addDelta("gradients")

    local adx = 1-dx
    local ady = 1-dy

    local g1value = {g1[1]*dx, g1[2]*dy}
    local g2value = {-g2[1]*adx, -g2[2]*ady}
    local g3value = {topRight and -g3[1]*adx or g3[1]*dx,  topRight and g3[2]*dy or -g3[2]*ady}

    debugUtil.addDelta("values")

    local g1di =self:fade(math.max(0, 0.7 - math.sqrt(dx*dx+dy*dy)))
    local g2di = self:fade(math.max(0, 0.7- math.sqrt(adx*adx + ady*ady)))
    local g3di = self:fade(math.max(0, 0.7-math.sqrt((topRight and adx*adx or dx*dx)+(topRight and dy*dy or ady*ady))))



    local g1influence = {g1di,g1di}--{topRight and adx or adx/2, topRight and ady/2 or ady}
    local g2influence = {g2di,g2di}--{topRight and dx/2 or dx, topRight and dy or dy/2}
    local g3influence = {g3di,g3di}--{topRight and dx/2 or adx/2, topRight and ady/2 or dy/2}
    debugUtil.addDelta("influences")

    local values = {g1value,g2value,g3value}
    local influences = {g1influence, g2influence, g3influence}
    local res = 0
    for i=1,3 do
      res = res + values[i][1]*influences[i][1] + values[i][2]*influences[i][2]
    end
    debugUtil.addDelta("sums")
    --res = g1value[1]*g1di + g2value[1]*g2di + g3value[1]*g3di


    return res

  end
  error("not yet")

end


local perm = {151,160,137,91,90,15,
  131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,
  190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,
  88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,
  77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,
  102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,
  135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,
  5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,
  223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,
  129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,
  251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,
  49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
  138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180,
  151,160,137,91,90,15,
  131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,
  190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,
  88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,
  77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,
  102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,
  135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,
  5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,
  223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,
  129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,
  251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,
  49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
  138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180}

local grad3= {{1,1,0},{-1,1,0},{1,-1,0},{-1,-1,0},
             {1,0,1},{-1,0,1},{1,0,-1},{-1,0,-1},
             {0,1,1},{0,-1,1},{0,1,-1},{0,-1,-1}}


function Noise:get2d(xin,yin)
  local n0,n1,n2

  local F2 = 0.5*(math.sqrt(3)-1)
  local s = (xin + yin)*F2
  local i = math.floor(xin+s)
  local j = math.floor(yin+s)

  local G2 = (3-math.sqrt(3))/6
  local t = (i+j)*G2
  local X0 = i-t
  local Y0 = j-t
  local x0 = xin - X0
  local y0 = yin- Y0

  local i1,j1
  if x0 > y0 then i1=1; j1=0; else i1=0; j1=1; end

  local x1 = x0-i1+G2
  local y1 = y0-j1+G2
  local x2 = x0-1+2*G2
  local y2 = y0-1+2*G2

  local ii = i % 256
  local jj = j % 256

  local gi0 = 1+perm[1+ii+perm[1+jj]] % 12
  local gi1 = 1+perm[1+ii+i1+perm[1+jj+j1]] % 12
  local gi2 = 1+perm[1+ii+1+perm[1+jj+1]] % 12

  local t0=0.5-x0*x0-y0*y0
  if t0 < 0 then n0 = 0 else t0=t0*t0; n0=t0*t0*(grad3[gi0][1]*x0 + grad3[gi0][2]*y0) end

  local t1=0.5-x1*x1-y1*y1
  if t1 < 0 then n1 = 0 else t1=t1*t1; n1=t1*t1*(grad3[gi1][1]*x1 + grad3[gi1][2]*y1) end

  local t2=0.5-x2*x2-y2*y2
  if t2 < 0 then n2 = 0 else t2=t2*t2; n2=t2*t2*(grad3[gi2][1]*x2 + grad3[gi2][2]*y2) end
  return 70 * (n0+n1+n2)

end

return Noise
