local points = require("point")

local M = {}
local Renderer = {}
M.Renderer = Renderer

function Renderer:new(blockSize, screenWidth, screenHeight)
  self.__index = self
  local newObj = {
    blockSize = blockSize,
    screenWidth = screenWidth,
    screenHeight = screenHeight
    }
  return setmetatable(newObj, self)
end

function Renderer:setHud(hud)
  self.hud = hud
end


function Renderer:render(gameArea, centerPos)
  -- centerPos = centerPos:round()
  self.centerPos = centerPos
  local screeHalf = points.new(self.screenWidth/2, self.screenHeight/2):round():assertInt()

  self.topLeftShift = (centerPos - screeHalf) -- :assertInt()


  self.batchIndex = 0
  for i, layer in ipairs(gameArea.layers) do
    self.renderIndex = i
    self:renderLayer(layer)
  end

  for i, entity in ipairs(gameArea.entities) do
    self:renderEntity(entity)
  end

  local hud = self.hud
  if hud then
    hud.size = points.new(love.graphics.getWidth(), love.graphics.getHeight())
    hud.position:set2(0,0)
    self.hud:draw()
  end


end

function Renderer:renderEntity(entity)
  local pos = entity.position - self.topLeftShift --entity.position:round() - self.topLeftShift:assertInt()

  local topLeft = pos
  local bottomRight = topLeft + points.new(entity.shapeWidth, entity.shapeHeight)

  if bottomRight.x < 0 or bottomRight.y < 0 or topLeft.x > self.screenWidth or topLeft.y > self.screenHeight then
    -- offscreen entity - no need for rendering
    return
  end

  local tex = textures[entity.material.textureId]
  assert(tex, "no texture id "..tostring(entity.material.textureId))
  local scaleX = entity.shapeWidth / tex:getWidth()
  local scaleY = entity.shapeHeight / tex:getHeight()
  --assert(pos.x==math.floor(pos.x))
  --assert(pos.y==math.floor(pos.y))
  love.graphics.draw(tex, pos.x, pos.y, nil, scaleX, scaleY)

  if entity.equipment then
    for equipment, v in pairs(entity.equipment) do
      local targetPos = pos + equipment.shift
      local size = equipment.targetSize
      local c = {love.graphics.getColor()}
      love.graphics.setColor(60,255,60,128)
      love.graphics.rectangle("fill", targetPos.x, targetPos.y, size.x, size.y)
      love.graphics.setColor(unpack(c))
    end
  end
end

-- iterator for (index, position) based on center and screen size
local function forBlockIndexAndPosition(center, length, blockSize)
  local startPixel = center - length/2 -- math.floor(length/2+0.5)
  local blockCount = math.ceil( length / blockSize)+1
  local startBlockIndex = math.floor(startPixel/blockSize)



  local currentIndexShift = 0
  return function()
    local i = currentIndexShift
    currentIndexShift  = currentIndexShift + 1

    if i >= blockCount then
      return nil
    else
      local blockIndex = startBlockIndex + i
      return blockIndex, blockIndex * blockSize - startPixel
    end
  end
end


function Renderer:renderChunk(chunk)
  if not chunk.viewCache then
    chunk.viewCache = {}
  end

  local batch = chunk.viewCache.batch
  local newBatch = not batch
  if not  batch or chunk.changed then
    if not batch then
      chunk.viewCache.batch = love.graphics.newSpriteBatch(self.atlasImage, chunk:getBlocksCount(), "stream")
      batch = chunk.viewCache.batch
    else
      if batch:getTexture()~= self.atlasImage then
        batch:setTexture(self.atlasImage)
      end
      --batch:clear()
    end


    local layer = chunk.layer
    local chunkSize = layer.chunkSize
    for dy=0, chunkSize-1 do
      for dx=0, chunkSize-1 do
        local block = chunk.blocks and chunk.blocks[1+dy*chunkSize+dx] or layer.defaultBlock
        local texture, r,g,b,a = self:getBlockRenderParams(block, self.renderIndex, layer.index)
        if dx == 0 or dy == 0  then
          --if dx==0 and chunk:getChunkX()%16==0 or dy==0 and chunk:getChunkY()%16==0 then
            --texture = textures["broken"]
          -- else
          --   r,g,b,a = 220,220,255,255
          -- end
        end

        batch:setColor(r, g, b, a)

        if newBatch then
          assert(0~=batch:add(self:getTextureQuad(texture), dx*BLOCK_SIZE, dy*BLOCK_SIZE))
        else
           assert(0~=batch:set(1+dy*chunkSize+dx, self:getTextureQuad(texture), dx*BLOCK_SIZE, dy*BLOCK_SIZE))
        end
      end
    end
    batch:flush()
    chunk.changed = false
  end
end

function Renderer:getBlockRenderParams(block, renderIndex, layerIndex)
  local texture =  textures[block.material.textureId];
  if not texture then
    texture = textures["broken"]
  end

  local durabilityRatio = block.durability / block.maxDurability
  local r = 64
  local g=0
  local b=0
  local a=255
  if durabilityRatio > 0 then
    local isSky = block.blockId == "core:sky"
    local showSky = renderIndex == 1 and isSky
    local bright = layerIndex == 0
    local brightness = (showSky) and 1 or (math.pow(0.6, layerIndex))
    local alpha = (isSky and not showSky) and 0 or 255

    local fullColor = brightness * 255
    local durabilityColor = fullColor * durabilityRatio

    r,g,b,a = fullColor, durabilityColor, durabilityColor,alpha
  end
  return texture, r,g,b,a
end

function Renderer:initAtlas()
  if not self.atlas then
    self.atlasEntrySize = self.blockSize
    self.atlasSize = 10
    self.atlas = love.image.newImageData(self.atlasEntrySize  * self.atlasSize, self.atlasEntrySize  * self.atlasSize)
    --self.atlasBatch = love.graphics.newSpriteBatch(love.graphics.newImage(self.atlas), 4*math.floor( self.screenHeight*self.screenWidth/self.blockSize/self.blockSize), "stream")
    self.atlasImage = love.graphics.newImage(self.atlas)
    self.textureIndices = {}
    self.quads = {}
    self.atlasUpdated = true
  end
end

function Renderer:getTextureQuad(texture)
  if not self.textureIndices.count then
    self.textureIndices.count = 0
  end
  local index = self.textureIndices[texture]
  local newTexture = not index
  if newTexture then
    index = self.textureIndices.count
    self.textureIndices[texture]  = index
    self.textureIndices.count = index + 1
  end



  local dx = self.atlasEntrySize * (index % self.atlasSize)
  local dy = self.atlasEntrySize * math.floor(index/self.atlasSize)
  if newTexture then
    self.atlasUpdated = true
    local texData = texture:getData()
    self.atlas:paste(texData, dx, dy, 0, 0, self.blockSize, self.blockSize)
    self.quads[index] = love.graphics.newQuad(dx,dy, self.blockSize, self.blockSize, self.atlas:getDimensions())
  end

  return self.quads[index]
end

function Renderer:renderLayer(layer)
  local renderIndex = self.renderIndex

  self:initAtlas()

  local batch = self.atlasBatch

  local level = layer
  local centerPos = self.centerPos
  local xOffset = -centerPos.x + math.floor(self.screenWidth/2)
  local yOffset = -centerPos.y + math.floor(self.screenHeight/2)
  local xFrac = xOffset % 1
  local yFrac = yOffset % 1

  

  xOffset = xOffset-xFrac
  yOffset = yOffset-yFrac

  local blockCounter = 0
  local tempColor = {love.graphics.getColor()}

  for _, chunk in pairs(layer.chunks) do
    self:renderChunk(chunk)
  end

  if self.atlasUpdated then
    self.atlasImage = love.graphics.newImage(self.atlas)
    for _,chunk in pairs(layer.chunks) do
      chunk.viewCache.batch:setTexture(self.atlasImage)
    end
    self.atlasUpdated = false
  end

  local canvasPadding = self.blockSize
  local canvasWidth  = self.screenWidth + canvasPadding
  local canvasHeight = self.screenHeight + canvasPadding
  if not self.canvas or self.canvas:getWidth() ~= canvasWidth  or self.canvas:getHeight() ~= canvasHeight then
    self.canvas = love.graphics.newCanvas(canvasWidth, canvasHeight)
  end

  self.canvas:renderTo(function()
    love.graphics.clear()
    for _, chunk in pairs(layer.chunks) do
      local chunkRenderX = chunk.fromPoint.x*BLOCK_SIZE + xOffset
      local chunkRenderY = chunk.fromPoint.y*BLOCK_SIZE + yOffset

      love.graphics.draw(chunk.viewCache.batch, chunkRenderX, chunkRenderY)
    end
  end)
  love.graphics.setBlendMode("alpha","premultiplied")
  love.graphics.draw(self.canvas, xFrac-1, yFrac-1)
  love.graphics.setBlendMode("alpha")

  --love.graphics.draw(self.atlasImage,0,0)


  love.graphics.setColor(unpack(tempColor))
  self.blocksRendered = blockCounter
end


return M
