local points = require("point")
local debugUtil = require("debug.debugUtil")

local NO_COLLISION = 1000
local BlockCollider = {}

function BlockCollider:new(layer)
  self.__index = self
  local instance = {
    layer = layer
  }
  return setmetatable(instance,self)
end

function BlockCollider:setLayer(layer)
  self.layer = layer
end


local function blockBorderLines(middleCoord, deltaCoord, isStart)
  local l = math.floor(middleCoord / BLOCK_SIZE)*BLOCK_SIZE
  local r = math.ceil(middleCoord / BLOCK_SIZE)*BLOCK_SIZE
  if (deltaCoord > 0) == isStart then
    return r
  else
    return l
  end
end

local otherField = {x="y",y='x'}

local function getCollisionRatio(fieldName, shiftedStart, shiftedEnd, entitySize, layer)
  local other = otherField[fieldName]
  local delta = shiftedEnd - shiftedStart
  if math.abs(delta[fieldName])==0 then
    return NO_COLLISION
  end

  local rowCheckDirection = points.new(delta.x > 0 and -1 or 1, delta.y > 0 and -1 or 1)
  local rowCheckSize = points.new(math.ceil(entitySize.x/BLOCK_SIZE), math.ceil(entitySize.y/BLOCK_SIZE))

  local l = blockBorderLines(shiftedStart[fieldName], delta[fieldName], true)
  local r = blockBorderLines(shiftedEnd[fieldName], delta[fieldName], false)
  if (r-l)*delta[fieldName] < 0 then
    return NO_COLLISION
  end
  local steps = math.floor( (r-l)/BLOCK_SIZE + 0.5)
  local stepSize = r > l and 1 or -1


  local blockNegative = delta[fieldName] < 0
  local blockShift = blockNegative and -1 or 0
  for shiftSteps = 0,steps,stepSize do

    local coord = l + BLOCK_SIZE*shiftSteps
    local otherCoord = shiftedStart[other] +(coord - shiftedStart[fieldName])/delta[fieldName]*delta[other]

    local blockCoord = math.floor(coord / BLOCK_SIZE+0.5) + blockShift
    local otherBlockCoord = math.floor(otherCoord / BLOCK_SIZE)

    local blockCoords = points.new()
    blockCoords[fieldName] = blockCoord
    blockCoords[other] = otherBlockCoord

    local impactBlocks = {}
    local impactBlockAlreadyAdded={}

    local function tryAddImpactBlock(impactCoord)
      local impactPosition = points.new()
      impactPosition[fieldName]=coord+blockShift*BLOCK_SIZE
      impactPosition[other]=impactCoord
      local block = layer:getBlockPx(impactPosition)
      if impactBlockAlreadyAdded[block] then
        return
      end

      if not block.passable then
        --debugUtil.trace(impactPosition, "impact Position!")
        impactBlockAlreadyAdded[block] = block
        table.insert(impactBlocks, block)
      end
    end



    local oldImpact = {}
    for indexInRowShift = 0, rowCheckSize[other]-1 do
      local changedCoord = blockCoords:clone()
      changedCoord[other] = changedCoord[other] + indexInRowShift * rowCheckDirection[other]
      local block = layer:getBlock(changedCoord.x, changedCoord.y)
      if not block.passable then
        table.insert(oldImpact, block)
      end
    end
    if #oldImpact > 0 then
    --  mobdebug.on()
      local d = true
    end

    local impactCoord = otherCoord

    while math.abs(impactCoord - otherCoord) < entitySize[other] do
      tryAddImpactBlock(impactCoord)
      impactCoord = impactCoord + rowCheckDirection[other]*BLOCK_SIZE
    end
    local endCoord = otherCoord + rowCheckDirection[other]*entitySize[other]
    local exclusiveEnd =  fmod(endCoord, BLOCK_SIZE) == 0 and rowCheckDirection[other] > 0
    if not exclusiveEnd then
      tryAddImpactBlock(endCoord)
    end


    if #impactBlocks > 0 then
      local pt = points.new()
      pt[fieldName] = coord
      pt[other] = otherCoord
      local ratio = (pt-shiftedStart):length()/delta:length()

      --return steps > 0 and shiftSteps/steps or 0
      return ratio, impactBlocks
    end


  end
  return NO_COLLISION
end





function BlockCollider:handle(entity)
  debugUtil.resetDelta()
  --if not self.layer then
  --  error("you forgot to set layer for block collider")
 -- end

  if not entity.lastStep then
    return
  end





  local start = entity.lastStep.position:clone()
  local endPoint = entity.position:clone()

  if start:equals(endPoint) then
    return
  end


  local noMoreReflections = false
  entity.lastStep.reflections = {}
  local safeCount = 1000
  debugUtil.addDelta("preparations")
  repeat
    safeCount = safeCount - 1
    if safeCount <= 0 then
      break
    end

    local delta = endPoint - start
    local collisionCornerShift = points.new(delta.x > 0 and entity.shapeWidth or 0, delta.y > 0 and entity.shapeHeight or 0)
    local shiftedStart = start + collisionCornerShift
    local shiftedEnd  = endPoint + collisionCornerShift

    local verticalLineX, impactBlocksX = getCollisionRatio("x",shiftedStart, shiftedEnd, entity:getSize(),self.layer)
    local horizontalLineY, impactBlocksY = getCollisionRatio("y",shiftedStart, shiftedEnd, entity:getSize(), self.layer)
    debugUtil.addDelta("collision ratio")
    if  verticalLineX == NO_COLLISION and horizontalLineY == NO_COLLISION then
      --if shouldBeCollizion then
        --print("NO COLLIZIONS (but should be) with "..tostring(endBlock.position).." pos"..tostring(endPoint))
      -- end
      noMoreReflections = true

    else
      --print(">>>>>>>>>COLLIDING!")

      local reflectCoord = verticalLineX < horizontalLineY and "x" or "y"
      local impactBlocks = verticalLineX < horizontalLineY and impactBlocksX or impactBlocksY
      if not impactBlocks then
        error("no impact blocks!")
        mobdebug.on()
        if verticalLineX < horizontalLineY then
          local verticalLineX, impactBlocksX = getCollisionRatio("x",shiftedStart, shiftedEnd, entity:getSize(),self.layer)
        else
          local horizontalLineY, impactBlocksY = getCollisionRatio("y",shiftedStart, shiftedEnd, entity:getSize(), self.layer)
        end

      end
      --debugUtil.trace(impactBlocks, "impactBlocks")

     -- mobdebug.on()
      if not pcall(function()
        table.sort(impactBlocks, function(a,b) return a.material.elasticity < b.material.elasticity end)
      end) then
        mobdebug.on()
        local debug = true
      end




      local totalBlockElasticity = 0

      for _, block in ipairs(impactBlocks) do
        totalBlockElasticity = totalBlockElasticity + block.material.elasticity
      end

      local avgBlockElasticity = totalBlockElasticity / #impactBlocks
      local resultElasticity = (avgBlockElasticity + entity.material.elasticity)/2



      local ratio = math.min(verticalLineX, horizontalLineY)
      local middlePoint = start + delta * ratio
      middlePoint[reflectCoord] = math.floor(middlePoint[reflectCoord]+0.5)

      table.insert(entity.lastStep.reflections, {coordName=reflectCoord, position=middlePoint})




      local realEndSpeed = entity.speed:clone()
      local originalEnergy = entity.mass * entity.speed[reflectCoord] ^ 2 / 2

      local timeEstimate = (endPoint - middlePoint):length()/entity.speed:length()

      local perBlock = originalEnergy / #impactBlocks
      local lossLess = 0
      local lossFull = 0



      local blockImpulse = 0
      local currentSpeed = entity.speed[reflectCoord]

      for _,block in ipairs(impactBlocks) do
        local blockCanConsume = block:getCanConsumeEnergy()
        if blockCanConsume >= perBlock or not block.destructible then
          lossLess = lossLess + perBlock
        else
          local afterConsume = perBlock - blockCanConsume
          if afterConsume <= block:getDurability() then
            block:setDurability(block:getDurability()-afterConsume)
            lossLess = lossLess + blockCanConsume
            lossFull = lossFull + afterConsume
          else
            local durabilityBefore = block:getDurability()
            lossFull = lossFull +  blockCanConsume + durabilityBefore
            local remainingEnergy = afterConsume - durabilityBefore
            block:setDurability(0)

            local currentSpeedSign = currentSpeed > 0 and 1 or -1

            local entityEnergy = currentSpeed * currentSpeed * entity.mass / 2
            entityEnergy = math.max(0,entityEnergy - durabilityBefore)
            currentSpeed = currentSpeedSign * math.sqrt(2*entityEnergy/entity.mass)

            local blockSpeed = currentSpeed  * 2 * entity.mass / (entity.mass + block.mass)
            currentSpeed = currentSpeed * (entity.mass - block.mass) / (entity.mass + block.mass)

            --local scalarSpeed = math.sqrt(remainingEnergy*2/block.mass)
            local speed = points.new()
            speed[reflectCoord] = blockSpeed
            --blockImpulse = blockImpulse +  block.mass *scalarSpeed
            if block.entity then
              block.entity.speed:set(speed)
            else
              block.entitySpeed = speed
            end
          end
        end
      end

      --mobdebug.on()

      lossLess = lossLess * resultElasticity

      local originalSpeedSign = entity.speed[reflectCoord] > 0 and 1 or -1

      local endSpeed
      if lossLess > 0 then
        endSpeed = -1 * originalSpeedSign * math.sqrt(lossLess * 2 / entity.mass)
      else
        --local speedEnergy = math.max(0, originalEnergy - lossFull)
        --local startImpulseSpeed= math.sqrt(speedEnergy * 2 / entity.mass)
        --local startImpulse = entity.mass * startImpulseSpeed

        --endSpeed = originalSpeedSign* (startImpulse - blockImpulse)/entity.mass
        endSpeed = currentSpeed
      end


      realEndSpeed[reflectCoord] = math.abs(endSpeed)  < 2 and 0 or endSpeed

      entity.speed:set(realEndSpeed)

      local reflectedDelta = realEndSpeed * timeEstimate
      local realEndPoint = middlePoint + reflectedDelta

      -- prepare for next iteration
      start = middlePoint
      endPoint = realEndPoint
      debugUtil.addDelta("collision resolution")
    end
  until noMoreReflections
  entity.position:set(endPoint)
  --print("y speed "..oldSpeed.."->"..entity.speed.y.." pos "..entity.lastStep.position.y..","..oldY.."->"..entity.position.y)

end







return BlockCollider
