local ffi = require("ffi")
ffi.cdef[[
   typedef struct {
     double x;
     double y;
   } point_t;
]]

local M = {}

local newPoint
local Point = {
   __add = function(a,b)
     return newPoint(a.x+b.x, a.y+b.y)
   end,

   __mul = function(pt, number)
     return newPoint(pt.x*number,pt.y*number)
   end,

   __sub = function(a,b)
     return newPoint(a.x-b.x, a.y-b.y)
   end,

   __tostring = function(self)
     return ""..tostring(self.x)..","..(self.y)
   end,

   __index = {
     set=function(self,other)
       self.x = other.x
       self.y = other.y
       return self
     end,

     set2=function(self,x,y)
       self.x = x
       self.y = y
       return self
     end,

     length=function(self)
       return math.sqrt(self.x*self.x+self.y*self.y)
     end,

     lengthSquared= function(self)
       return self.x*self.x + self.y*self.y
     end,

     scaledTo = function(self,newLength)
       local originalLength = self:length()
       if originalLength  == 0 then
         return newPoint(newLength,0)
       else
         local ratio = newLength/originalLength
         return self * ratio
       end
     end,

     clone=function(self)
       return newPoint(self.x,self.y)
     end,

     cappedLength=function(self,maxLength)
       if maxLength  < self:length() then
         return self:scaledTo(maxLength)
       else
         return self:clone()
       end
     end,


     modInsideRect=function(self,topLeft, bottomRight)
       local newX = self.x
       local newY = self.y
       local w = bottomRight.x - topLeft.x
       local h = bottomRight.y - topLeft.y

       while newX < topLeft.x do
         newX = newX + w
       end
       while newY < topLeft.y do
         newY = newY + h
       end
       while newX > bottomRight.x do
         newX = newX - w
       end
       while newY > bottomRight.y do
         newY = newY - h
       end
       return newPoint(newX,newY)

     end,

     round = function(self)
       local pos = newPoint(math.floor(self.x+0.5),math.floor(self.y+0.5))
       return pos
     end,

     floor = function(self)
       return newPoint(math.floor(self.x),math.floor(self.y))
     end,

     ceil = function(self)
       return newPoint(math.ceil(self.x),math.ceil(self.y))
     end,

     toLongerInt = function(self)
       return newPoint(self.x > 0 and math.ceil(self.x) or math.floor(self.x), self.y>0 and math.ceil(self.y) or math.floor(self.y))
     end,

     assertInt = function(self)
       assert(self.x==math.floor(self.x))
       assert(self.y==math.floor(self.y))
       return self
     end,


     swapped = function(self)
       return newPoint(self.y, self.x)
     end,

     equals = function(self,other)
       return self.x==other.x and self.y==other.y
     end


   }
}




newPoint = ffi.metatype("point_t", Point)


function M.createPointCollection(coordsList)
  local result = {}
  for i = 1,#coordsList,2 do
    local x = coordsList[i]
    local y = coordsList[i+1]
    table.insert(result,newPoint(x,y))
  end
  return result
end


M.Point = Point
M.deltas = {}
M.deltas.up = newPoint(0,-1)
M.deltas.down = newPoint(0,1)
M.deltas.left = newPoint(-1,0)
M.deltas.right = newPoint(1,0)
M.new = newPoint


return M
