
local threadCallbacks = require("threads.ThreadCallBacks").getInstance()

local M = {}
function M.extractExtension(path)
  local nameWithExtension = M.extract(path):reverse()
  local extensionEnd = nameWithExtension:find(".",nil, true)-1
  local extensionInverted = nameWithExtension:sub(0,extensionEnd)
  local extension = extensionInverted:reverse()
  -- print("path "..path.."=>"..nameWithExtension.."=>("..fileNameStart..")"..name)
  return extension
end
function M.extractName(path)
    local nameWithExtension = M.extract(path):reverse()
    local fileNameStart = nameWithExtension:find(".",nil, true)+1
    local nameInverted = nameWithExtension:sub(fileNameStart)
    local name = nameInverted:reverse()
    -- print("path "..path.."=>"..nameWithExtension.."=>("..fileNameStart..")"..name)
    return name
end

function M.extract(path)
  local fileNameBack  = path:reverse()
  local fileNameEnd = fileNameBack:find("/")
  if not fileNameEnd then
    return path
  else
    local nameInverted = fileNameBack:sub(1, fileNameEnd-1)
    local name = nameInverted:reverse()
    return name
  end
end

function M.forChildPaths(parentPath, handler, filter)
  if type(filter)=="string" then
    local originalFilter = filter
    filter = function(path)
      return path:match(originalFilter)~=nil;
    end
  elseif type(filter)~="function" then
    filter = function() return true; end
  end
  for i, childName in ipairs(love.filesystem.getDirectoryItems(parentPath)) do

    local childPath = parentPath.."/"..childName
    if filter(childName) then
      handler(childPath)
    end
  end
end


function M.getModsRoot()
  return "assets/mods";
end

function M.getSavesRoot()
  return "assets/saves";
end


local ensuredChunksDirExists = false
function M.getChunksRoot()
  local result =  M.getSavesRoot().."/chunks"
  if not ensuredChunksDirExists then
    love.filesystem.createDirectory(result)
    ensuredChunksDirExists = true
  end
  return result
end

local ensuredRegionsDirExists = false
function M.getRegionsRoot()
  local result = M.getSavesRoot().."/regions"
  if not ensuredChunksDirExists then
    love.filesystem.createDirectory(result)
    ensuredRegionsDirExists = true
  end
  return result
end

function M.pathToId(path)
  local prefix = M.getModsRoot()
  if path:find(prefix) then
    path = path:sub(prefix:len()+1)
  end
  local suffix = M.extractExtension(path)
  if suffix then
    path = path:sub(1, path:len()-suffix:len())
  end
  path = path:gsub("/",":")
  return path
end

function M.getChunkFileNameById(chunkId)
  return M.getChunksRoot().."/"..chunkId..".txt"
end


local _loadThread
local function getLoadThead()
  if not _loadThread then
    _loadThread = love.thread.newThread("threads/loadThread.lua")
    _loadThread:start()
  end
  if _loadThread:getError() then
     io.stderr:write("Load thread error: "..tostring(_loadThread:getError()).."\n")
  end
  if not _loadThread:isRunning() then
     io.stderr:write("Load thread not running!\n")
  end
  return _loadThread
end

local _saveThread
local function getSaveThread()
  if not _saveThread then
    _saveThread = love.thread.newThread("threads/saveThread.lua")
    _saveThread:start()
  end
  if _saveThread:getError() then
    io.stderr:write("save thread error: "..tostring(_saveThread:getError()).."\n")
  end
  if not _saveThread:isRunning() then
    io.stderr:write("save thread not running!\n")
  end
  return _saveThread
end

local loadThread = getLoadThead()
local loadParamsChannel = love.thread.getChannel("loadFileChannelParams")
local saveParamsChannel = love.thread.getChannel("saveFileChannelParams")

local saveThread = getSaveThread()


function M.loadFileAsync(path, callBack)
  loadParamsChannel:push({path, threadCallbacks:registerCallBack(callBack)})
end


function M.saveFileAsync(path, data, dataSize, callBack)
  if math.random()<0.0001 and saveParamsChannel:getCount() > 10000 then
    print("BAD THING channel queue length = "..saveParamsChannel:getCount())
  end
  saveParamsChannel:push({path, data, dataSize, threadCallbacks:registerCallBack(callBack)})
end

local rwFileChannelParamsName = "rwFileChannelParams"
local rwFileChannel = nil
local rwFileThread = nil
local function getRwFileChannel()
  if not rwFileChannel then
    rwFileChannel = love.thread.getChannel(rwFileChannelParamsName)
    rwFileThread = love.thread.newThread("threads/rwFileThread.lua")
    rwFileThread:start(rwFileChannelParamsName)
  end
  return rwFileChannel
end

function M.doRwTask(taskName, taskArgs, callBack)
  local channel = getRwFileChannel()
  if rwFileThread:getError() then error(rwFileThread:getError()) end
  local params = {}
  local i =1
  local function addParam(param) params[i] = param; i=i+1 end
  addParam(taskName)


  for i=1, #taskArgs do addParam(taskArgs[i]) end

  addParam(threadCallbacks:registerCallBack(callBack))
  assert(#params >=3)
  channel:push(params)
end

function M.openFileAsync(path, callBack)
  M.doRwTask("open", {path}, callBack)
end

function M.closeFileAsync(fileHandle, callBack)
  assert(fileHandle)
  M.doRwTask("close", {fileHandle}, callBack)
end

function M.readFileAsync(fileHandle, offset, length, callBack)
  assert(fileHandle)
  assert(length)
  --print("Offset="..offset.." lenght="..length)
  M.doRwTask("read", {fileHandle, offset,length}, callBack)
end

function M.writeFileAsync(fileHandle, offset, dataPtr, length, callBack)
  assert(fileHandle)
  M.doRwTask("write", {fileHandle, offset,dataPtr, length}, callBack)
end

return M
